// ap_pcfg_openfile()
/*
 *
 * This module implements a set of Include commands for Apache 1.3.x.
 * by Mark Murawski
 *
 * Include a file or all files in a directory.  If there are syntax
 * errors, apache will continue loading peacefully.
 *
 * See README for usage
 * To build: have the apache axps program functional, type make, and enjoy.
 *
 * You will need something like the following in your apache config
 * LoadModule safeinclude_module /usr/lib/apache/1.3/mod_safeinclude.so
 *
 * Based on sql-include by Marcin Orlowski
 *
 */

#define MYNAME     "SafeInclude"
#define MYVERSION  "0.2"

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_main.h"
#include "http_log.h"
#include "util_script.h"
#include "http_conf_globals.h"

#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>

// #define       DEBUG

#ifdef DEBUG
#define D(x,y) fprintf(stderr, x, y);
#else
#define D(x,y)
#endif


#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

module MODULE_VAR_EXPORT safeinclude_module;

#define VERBOSE_LEVEL_QUIET             0
#define VERBOSE_LEVEL_BASIC             1
#define VERBOSE_LEVEL_DEBUG             2

//  SQL_VHost module config structure

#define SQL_SERVER_LEN          256
#define SQL_USER_LEN             80
#define SQL_PASSWORD_LEN         80
#define SQL_DB_LEN               80

typedef struct {
   int     AbortOnError;
   int     Verbose;
} server_conf;    

///////////////////////////////////////////
//         Start Your Engines            //
///////////////////////////////////////////

// Make it known we loaded
//
static void init_module (server_rec *server, struct pool *p)
{
  ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, server, "%s: v%s by Mark Murawski <kobaz@kobaz.net>", MYNAME, MYVERSION);
}

// read in config data
//
int GetLine (char *dest, size_t dest_size, char **src_ptr)
{
  int i;
  char *src = *src_ptr;

  // let's 'read' the src data till closest LF or \0
  for (i = 0; i < dest_size; i++, src++) {
    switch (*src) {
     case '\0':
      dest[i] = '\0';
      D("'%s'\n", dest)
      *src_ptr = src;

      if (i == 0)
        return 0;              // EOF
      else
        return 1;              // we need to do so, otherwise last line

      // wouldn't be processed at all
      break;
      
     case '\n':
      dest[i] = '\0';
      D("'%s'\n", dest);

      // storing source pointer back for further
      *src_ptr = src+1;

      //  reads
      return 1;                      // We'll be back here...
      break;
      
     default:
      dest[i] = *src;
      break;
    }
  }

  D("'%s'\n", dest);
  return 0;  // shouldn't happen
}

// include holds plain ascii taken from SQL base.
int ProcessEntry (server_rec *s, pool *p, pool *ptemp, char *data, char *file)
{
  cmd_parms    parms = {NULL, 0, -1, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
  configfile_t *new_cfg;
  const char   *errmsg;

  // working data pointer - we need to remember where
  char *record_ptr = (char *) data;
  
  // let's find our config data
  server_conf *conf = (server_conf *) ap_get_module_config(s->module_config, &safeinclude_module);
  
  if (conf->Verbose >= VERBOSE_LEVEL_DEBUG)
    ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, s, "%s: Processing config data: %s",  MYNAME, data);

  new_cfg = ap_palloc(p, sizeof(*new_cfg));
  new_cfg->getstr = (void *) GetLine;      // our getstring function
  new_cfg->param = &record_ptr;            // data we gonna parse
  new_cfg->line_number = 0;
  
  parms.pool = p;
  parms.temp_pool = ptemp;
  parms.server = s;
  parms.override = (RSRC_CONF | OR_ALL) & ~(OR_AUTHCFG | OR_LIMIT);       // probably useless here
  parms.config_file = new_cfg;
  
  // this uses Apache core config file parser. All we have to do is to feed it
  // with the data, which is done using GetLine()
  errmsg = ap_srm_command_loop(&parms, s->lookup_defaults);

  if (errmsg) {
    ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s,
      "%s: Syntax error in file %s, line %d", 
      MYNAME, file, parms.config_file->line_number);

    ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "%s: %s", MYNAME, errmsg);
    return 0;
  }
  
  return 1;
}

/*
 * Include a file
 */
static const char *cmd_safe_include (cmd_parms *cmd, void *dummy, char *arg)
{
  FILE *file = NULL;
  char *data = NULL;
  struct stat stat;
  pool *p = ap_make_sub_pool(cmd->pool);

  // let's find our config data
  server_rec *s = cmd->server;
  server_conf *conf = (server_conf *) ap_get_module_config(s->module_config, &safeinclude_module);
  
  if (conf->Verbose >= VERBOSE_LEVEL_BASIC)
    ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, s, "%s: Attempting to Include %s", MYNAME, arg);
  
  if (!(file = ap_pfopen(cmd->pool, arg, "r"))) {
    ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "%s: Can't open '%s' (%s)", MYNAME, arg, strerror(errno));
    goto error;
  }

  if (fstat(fileno(file), &stat) == -1) {
    ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "%s: Can't stat '%s' (%s)", MYNAME, arg, strerror(errno));
    goto error;
  }

  data = ap_palloc(cmd->pool, (stat.st_size + 1));
  fread(data, sizeof(char), stat.st_size, file);
  data[stat.st_size] = 0; // make this a proper string

  // process this file
  if ((ProcessEntry(cmd->server, cmd->pool, cmd->temp_pool, data, arg) == 0))
    goto error;

  if (conf->Verbose >= VERBOSE_LEVEL_BASIC)
    ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, s, "%s: Loaded: %s", MYNAME, arg);

  return NULL;

 error:
  ap_destroy_pool(p);

  if (conf->AbortOnError == TRUE)
    return "Include Failed";

  return NULL;
}

static const char *cmd_safe_include_dir (cmd_parms *cmd, void *dummy, char *arg) {
  DIR *dir;
  struct dirent *dirinfo;
  struct stat stat;
  pool *p = ap_make_sub_pool(cmd->pool);
  server_rec *s = cmd->server;
  char *path;

  // let's find our config data
  server_conf *conf = (server_conf *) ap_get_module_config(s->module_config, &safeinclude_module);

  if (conf->Verbose >= VERBOSE_LEVEL_BASIC)
    ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, s, "%s: Searching dir for files: %s",  MYNAME, arg);

  if (!(dir = ap_popendir(p, arg))) {
    ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "%s: Can't read directory '%s' (%s)", MYNAME, arg, strerror(errno));
    goto error;
  }

  while ((dirinfo = readdir(dir))) {
    if (!strcmp(dirinfo->d_name, "."))
      continue;
    if (!strcmp(dirinfo->d_name, ".."))
      continue;
    
    path = ap_pstrcat(p, arg, "/", dirinfo->d_name, NULL);
    if (lstat(path, &stat) == -1) {
      ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "%s: Can't stat '%s/%s' (%s)", MYNAME, arg, dirinfo->d_name, strerror(errno));
      goto error;
    }

    if (!S_ISREG(stat.st_mode))
      continue;

    cmd_safe_include(cmd, dummy, path);
    errno = 0;
  }

  if (errno)
    ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "%s: Failed while reading directory '%s' (%s)", MYNAME, arg, strerror(errno));

 error:
  ap_destroy_pool(p);

  if (conf->AbortOnError == TRUE)
    return "Include Failed";

  return NULL;
}

// sets AbortOnError switch 
static const char *cmd_safe_abortonerror (cmd_parms *cmd, void *dummy, int arg)
{
  server_rec *s = cmd->server;
  server_conf *conf = (server_conf *) ap_get_module_config(s->module_config, &safeinclude_module);
  
  conf->AbortOnError = arg;
  return NULL;
}

// sets Verbose switch 
static const char *cmd_safe_verbose (cmd_parms *cmd, void *dummy, char *arg)
{
  server_rec *s = cmd->server;
  server_conf *conf = (server_conf *) ap_get_module_config(s->module_config, &safeinclude_module);
  
  conf->Verbose = atoi(arg);

  if (conf->Verbose < VERBOSE_LEVEL_QUIET)
    conf->Verbose = VERBOSE_LEVEL_QUIET;
  else {
    if (conf->Verbose > VERBOSE_LEVEL_DEBUG)
      conf->Verbose = VERBOSE_LEVEL_DEBUG;
  }  

  return NULL;
}

/*
** Let's allocate module config struct and fill with defaults
*/
static void *create_safe_config (pool *p, server_rec *s)
{
  server_conf *c = (server_conf *) ap_pcalloc(p, sizeof(server_conf));
  
  if (c) {
    // setting up defaults...
    c->AbortOnError = FALSE;
    c->Verbose = VERBOSE_LEVEL_BASIC;
  }
  
  return c;
}

/*
** Hare are the config directives our module implements
*/
static const command_rec safe_cmds[] =
{
  { "Safe_Include", cmd_safe_include, NULL, RSRC_CONF, TAKE1,
    "Any filename. Data in it will be passed to Apache. Must be a full path." },

  { "Safe_IncludeDir", cmd_safe_include_dir, NULL, RSRC_CONF, TAKE1,
    "Any directory. Files in it will be passed to Apache. Must be a full path." },

  { "Safe_AbortOnError", cmd_safe_abortonerror, NULL, RSRC_CONF, FLAG,
    "Process will be aborted on syntax error. Default 'Off'" },

  { "Safe_Verbose", cmd_safe_verbose, NULL, RSRC_CONF, TAKE1,
    "0,1,2 - level of module talkative. 0=errors only, 1=show loading info, 2=speak up baby." },
  
  { NULL }
};

module MODULE_VAR_EXPORT safeinclude_module =
{
    STANDARD_MODULE_STUFF,
    init_module,                /* initializer */
    NULL,                       /* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    create_safe_config,         /* server config */
    NULL,                       /* merge server config */
    safe_cmds,                  /* command table */
    NULL,                       /* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL                        /* post read-request */
};
